# 5 - Python's Class Methods
"""
    Class Methods are different than Instance Methods.
    Class Methods don't require an instance of the Class to be called.
    They can be called by the Class like Classname.methodname().
    You define a Class Method using @classmethod.
    The first argument is 'cls'; which is the Class where it was defined.
"""

"""
   Class methods are commonly used as factory methods.
   Factory Method - The main motivation behind the Factory Method Design Pattern is to enhance loose coupling in code through the creation of an abstract class that will be used to create different types of objects that share some common attributes and functionality.

    In this example, we use Factory Methods to create three types of Persons: Designer, Developer, and Tester.
   
"""

class Person:
    number_of_class_instances = 0

    def __init__(self, name, team, role):
        self.name = name
        self.team = team
        self.role = role
        Person.number_of_class_instances += 1

    def person_info(self):
        return f'Team: "{self.team}" Name: "{self.name}"'
    
    def __str__(self):
        return f'Team: "{self.team}" Name: "{self.name}" Role: "{self.role}"'
    
    #Class Method that creates a Designer.
    @classmethod
    def designer(cls, name, team):
        return cls(name, team, "Designer")
    
    #Class Method that creates a Developer.
    @classmethod
    def developer(cls, name, team):
        return cls(name, team, "Developer")
    
    #Class Method that creates a Tester.
    @classmethod
    def tester(cls, name, team):
        return cls(name, team, "Tester")
       
our_designer = Person.designer("John", "Team 1")
our_developer = Person.developer("Amy", "Team 1")
our_tester = Person.tester("Sherry", "Team 1")

for x in (our_designer, our_developer, our_tester):
    print(x.person_info())